"""Test utilities."""

from __future__ import annotations

import ast
import functools
import sys
from collections.abc import Callable
from typing import cast


def require_at_least(
    minor: int,
) -> Callable[[Callable[..., None]], Callable[..., None]]:
    """Require the minimum minor version of Python 3 to run the test.

    Args:
        minor: Minimum minor version (inclusive) that the test case supports.

    Returns:
        A decorator function to wrap the test case function.
    """

    def decorator(fn: Callable[..., None]) -> Callable[..., None]:
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            if sys.version_info.minor < minor:
                return
            fn(*args, **kwargs)

        return wrapper

    return decorator


def require_at_most(
    minor: int,
) -> Callable[[Callable[..., None]], Callable[..., None]]:
    """Require the maximum minor version of Python 3 to run the test.

    Args:
        minor: Maximum minor version (inclusive) that the test case supports.

    Returns:
        A decorator function to wrap the test case function.
    """

    def decorator(fn: Callable[..., None]) -> Callable[..., None]:
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            if sys.version_info.minor > minor:
                return
            fn(*args, **kwargs)

        return wrapper

    return decorator


def ast_equal(observed: ast.AST, expected: ast.AST) -> bool:
    """Checks the equality between two ASTs.

    This function checks if `observed` contains at least the same subtree with
    `expected`. If `observed` has some extra branches that `expected` does not cover,
    it is ignored.

    Args:
        observed: An AST to check.
        expected: The expected AST.

    Returns:
        True if observed and expected represent the same AST, False otherwise.
    """
    try:
        assert type(observed) is type(expected)

        for k, ve in vars(expected).items():
            if k in {"col_offset", "end_col_offset", "end_lineno", "kind", "lineno"}:
                continue

            vo = getattr(observed, k)  # May cause AttributeError.

            if isinstance(ve, ast.AST):
                assert ast_equal(cast(ast.AST, vo), ve)
            elif isinstance(ve, list):
                vo = cast(list, vo)
                assert len(vo) == len(ve)
                assert all(
                    ast_equal(cast(ast.AST, co), cast(ast.AST, ce))
                    for co, ce in zip(vo, ve)
                )
            else:
                assert type(vo) is type(ve)
                assert vo == ve

    except (AssertionError, AttributeError):
        return False

    return True


def assert_ast_equal(observed: ast.AST, expected: ast.AST) -> None:
    """Asserts the equality between two ASTs.

    Args:
        observed: An AST to compare.
        expected: Another AST.

    Raises:
        AssertionError: observed and expected represent different ASTs.
    """
    if sys.version_info.minor >= 9:
        assert ast_equal(
            observed, expected
        ), f"""\
AST does not match.
observed={ast.dump(observed, indent=4)}
expected={ast.dump(expected, indent=4)}
"""
    else:
        assert ast_equal(
            observed, expected
        ), f"""\
AST does not match.
observed={ast.dump(observed)}
expected={ast.dump(expected)}
"""



"""Parsing utilities."""

from __future__ import annotations

import ast
import inspect
import textwrap
from collections.abc import Callable
from typing import Any

import dill  # type: ignore[import]

from latexify import exceptions


def parse_function(fn: Callable[..., Any]) -> ast.Module:
    """Parses given function.

    Args:
        fn: Target function.

    Returns:
        AST tree representing `fn`.
    """
    try:
        source = inspect.getsource(fn)
    except Exception:
        # Maybe running on console.
        source = dill.source.getsource(fn)

    # Remove extra indentation so that ast.parse runs correctly.
    source = textwrap.dedent(source)

    tree = ast.parse(source)
    if not tree.body or not isinstance(tree.body[0], ast.FunctionDef):
        raise exceptions.LatexifySyntaxError("Not a function.")

    return tree



"""Definition of the Config class."""

from __future__ import annotations

import dataclasses
from typing import Any


@dataclasses.dataclass(frozen=True)
class Config:
    """Configurations to control the behavior of latexify.

    Attributes:
        expand_functions: If set, the names of the functions to expand.
        identifiers: If set, the mapping to replace identifier names in the
            function. Keys are the original names of the identifiers,
            and corresponding values are the replacements.
            Both keys and values have to represent valid Python identifiers:
            ^[A-Za-z_][A-Za-z0-9_]*$
        prefixes: Prefixes of identifiers to trim. E.g., if "foo.bar" in prefixes, all
            identifiers with the form "foo.bar.suffix" will be replaced to "suffix"
        reduce_assignments: If True, assignment statements are used to synthesize
            the final expression.
        use_math_symbols: Whether to convert identifiers with a math symbol surface
            (e.g., "alpha") to the LaTeX symbol (e.g., "\\alpha").
        use_set_symbols: Whether to use set symbols or not.
        use_signature: Whether to add the function signature before the expression
            or not.
    """

    expand_functions: set[str] | None
    identifiers: dict[str, str] | None
    prefixes: set[str] | None
    reduce_assignments: bool
    use_math_symbols: bool
    use_set_symbols: bool
    use_signature: bool

    def merge(self, *, config: Config | None = None, **kwargs) -> Config:
        """Merge configuration based on old configuration and field values.

        Args:
            config: If None, the merged one will merge defaults and field values,
                instead of merging old configuration and field values.
            **kwargs: Members to modify. This value precedes both self and config.

        Returns:
            A new Config object
        """

        def merge_field(name: str) -> Any:
            # Precedence: kwargs -> config -> self
            arg = kwargs.get(name)
            if arg is None:
                if config is not None:
                    arg = getattr(config, name)
                else:
                    arg = getattr(self, name)
            return arg

        return Config(**{f.name: merge_field(f.name) for f in dataclasses.fields(self)})

    @staticmethod
    def defaults() -> Config:
        """Generates a Config with default values.

        Returns:
            A new Config with default values
        """
        return Config(
            expand_functions=None,
            identifiers=None,
            prefixes=None,
            reduce_assignments=False,
            use_math_symbols=False,
            use_set_symbols=False,
            use_signature=True,
        )



"""Transformer to replace user symbols."""

from __future__ import annotations

import ast
import re
import sys
from typing import ClassVar, cast


class IdentifierReplacer(ast.NodeTransformer):
    """NodeTransformer to replace identifier names.

    This class defines a rule to replace identifiers in AST with specified names.

    Example:
        def foo(bar):
            return baz

        IdentifierReplacer({"foo": "x", "bar": "y", "baz": "z"}) will modify the AST of
        the function above to below:

        def x(y):
            return z
    """

    _IDENTIFIER_PATTERN: ClassVar[re.Pattern] = re.compile(r"^[A-Za-z_][A-Za-z0-9_]*$")

    def __init__(self, mapping: dict[str, str]):
        """Initializer.

        Args:
            mapping: User defined mapping of names. Keys are the original names of the
                identifiers, and corresponding values are the replacements.
                Both keys and values have to represent valid Python identifiers:
                ^[A-Za-z_][A-Za-z0-9_]*$
        """
        self._mapping = mapping

        for k, v in self._mapping.items():
            if not self._IDENTIFIER_PATTERN.match(k):
                raise ValueError(f"'{k}' is not an identifier name.")
            if not self._IDENTIFIER_PATTERN.match(v):
                raise ValueError(f"'{v}' is not an identifier name.")

    def _replace_args(self, args: list[ast.arg]) -> list[ast.arg]:
        """Helper function to replace arg names."""
        return [ast.arg(arg=self._mapping.get(a.arg, a.arg)) for a in args]

    def visit_FunctionDef(self, node: ast.FunctionDef) -> ast.FunctionDef:
        """Visit a FunctionDef node."""
        visited = cast(ast.FunctionDef, super().generic_visit(node))

        if sys.version_info.minor < 8:
            args = ast.arguments(
                args=self._replace_args(visited.args.args),
                kwonlyargs=self._replace_args(visited.args.kwonlyargs),
                kw_defaults=visited.args.kw_defaults,
                defaults=visited.args.defaults,
            )
        else:
            args = ast.arguments(
                posonlyargs=self._replace_args(visited.args.posonlyargs),  # from 3.8
                args=self._replace_args(visited.args.args),
                kwonlyargs=self._replace_args(visited.args.kwonlyargs),
                kw_defaults=visited.args.kw_defaults,
                defaults=visited.args.defaults,
            )

        return ast.FunctionDef(
            name=self._mapping.get(visited.name, visited.name),
            args=args,
            body=visited.body,
            decorator_list=visited.decorator_list,
        )

    def visit_Name(self, node: ast.Name) -> ast.Name:
        """Visit a Name node."""
        return ast.Name(
            id=self._mapping.get(node.id, node.id),
            ctx=node.ctx,
        )



"""NodeTransformer to reduce assigned expressions."""

from __future__ import annotations

import ast
from typing import Any

from latexify import exceptions


class AssignmentReducer(ast.NodeTransformer):
    """NodeTransformer to reduce assigned expressions.

    This class replaces a functions with multiple assignments to a function with only
    single return.

    Example:
        def f(x):
            y = 2 + x
            z = 3 * y
            return 4 + z

        AssignmentReducer modifies the function above to below:

        def f(x):
            return 4 + 3 * (2 + x)
    """

    _assignments: dict[str, ast.expr] | None = None

    # TODO(odashi):
    # Currently, this function does not care much about some expressions, e.g.,
    # comprehensions or lambdas, which introduces inner scopes.
    # It may cause some mistakes in the resulting AST.
    def visit_FunctionDef(self, node: ast.FunctionDef) -> Any:
        """Visit a FunctionDef node."""
        # Push stack
        parent_assignments = self._assignments
        self._assignments = {}

        for child in node.body[:-1]:
            if not isinstance(child, ast.Assign):
                raise exceptions.LatexifyNotSupportedError(
                    "AssignmentReducer supports only Assign nodes, "
                    f"but got: {type(child).__name__}"
                )

            value = self.visit(child.value)

            for target in child.targets:
                if not isinstance(target, ast.Name):
                    raise exceptions.LatexifyNotSupportedError(
                        "AssignmentReducer does not recognize list/tuple "
                        "decomposition."
                    )
                self._assignments[target.id] = value

        return_original = node.body[-1]

        if not isinstance(return_original, (ast.Return, ast.If)):
            raise exceptions.LatexifySyntaxError(
                f"Unsupported last statement: {type(return_original).__name__}"
            )

        return_transformed = self.visit(return_original)

        # Pop stack
        self._assignments = parent_assignments

        return ast.FunctionDef(
            name=node.name,
            args=node.args,
            body=[return_transformed],
            decorator_list=node.decorator_list,
        )

    def visit_Name(self, node: ast.Name) -> Any:
        """Visit a Name node."""
        if self._assignments is not None:
            return self._assignments.get(node.id, node)

        return node



from __future__ import annotations

import ast
import functools
from collections.abc import Callable

from latexify import ast_utils, exceptions


# TODO(ZibingZhang): handle mutually recursive function expansions
class FunctionExpander(ast.NodeTransformer):
    """NodeTransformer to expand functions.

    This class replaces function calls with an expanded form.

    Example:
        def f(x, y):
            return hypot(x, y)

        FunctionExpander({"hypot"}) will modify the AST of the function above to below:

        def f(x, y):
            return sqrt(x**2, y**2)
    """

    def __init__(self, functions: set[str]) -> None:
        self._functions = functions

    def visit_Call(self, node: ast.Call) -> ast.AST:
        """Visit a Call node."""
        func_name = ast_utils.extract_function_name_or_none(node)
        if (
            func_name is not None
            and func_name in self._functions
            and func_name in _FUNCTION_EXPANDERS
        ):
            return _FUNCTION_EXPANDERS[func_name](self, node)

        kwargs = {
            "func": self.visit(node.func),
            "args": [self.visit(x) for x in node.args],
        }

        if hasattr(node, "keywords"):
            kwargs["keywords"] = [
                ast.keyword(arg=x.arg, value=self.visit(x.value)) for x in node.keywords
            ]

        return ast.Call(**kwargs)


def _atan2_expander(function_expander: FunctionExpander, node: ast.Call) -> ast.AST:
    _check_num_args(node, 2)
    return ast.Call(
        func=ast.Name(id="atan", ctx=ast.Load()),
        args=[
            ast.BinOp(
                left=function_expander.visit(node.args[0]),
                op=ast.Div(),
                right=function_expander.visit(node.args[1]),
            )
        ],
    )


def _exp_expander(function_expander: FunctionExpander, node: ast.Call) -> ast.AST:
    _check_num_args(node, 1)
    return ast.BinOp(
        left=ast.Name(id="e", ctx=ast.Load()),
        op=ast.Pow(),
        right=function_expander.visit(node.args[0]),
    )


def _exp2_expander(function_expander: FunctionExpander, node: ast.Call) -> ast.AST:
    _check_num_args(node, 1)
    return ast.BinOp(
        left=ast_utils.make_constant(2),
        op=ast.Pow(),
        right=function_expander.visit(node.args[0]),
    )


def _expm1_expander(function_expander: FunctionExpander, node: ast.Call) -> ast.AST:
    _check_num_args(node, 1)
    return ast.BinOp(
        left=function_expander.visit(
            ast.Call(
                func=ast.Name(id="exp", ctx=ast.Load()),
                args=[node.args[0]],
            )
        ),
        op=ast.Sub(),
        right=ast_utils.make_constant(1),
    )


def _hypot_expander(function_expander: FunctionExpander, node: ast.Call) -> ast.AST:
    if not node.args:
        return ast_utils.make_constant(0)

    args = [
        ast.BinOp(
            left=function_expander.visit(arg),
            op=ast.Pow(),
            right=ast_utils.make_constant(2),
        )
        for arg in node.args
    ]

    args_reduced = functools.reduce(
        lambda a, b: ast.BinOp(left=a, op=ast.Add(), right=b), args
    )
    return ast.Call(
        func=ast.Name(id="sqrt", ctx=ast.Load()),
        args=[args_reduced],
    )


def _log1p_expander(function_expander: FunctionExpander, node: ast.Call) -> ast.AST:
    _check_num_args(node, 1)
    return ast.Call(
        func=ast.Name(id="log", ctx=ast.Load()),
        args=[
            ast.BinOp(
                left=ast_utils.make_constant(1),
                op=ast.Add(),
                right=function_expander.visit(node.args[0]),
            )
        ],
    )


def _pow_expander(function_expander: FunctionExpander, node: ast.Call) -> ast.AST:
    _check_num_args(node, 2)
    return ast.BinOp(
        left=function_expander.visit(node.args[0]),
        op=ast.Pow(),
        right=function_expander.visit(node.args[1]),
    )


def _check_num_args(node: ast.Call, nargs: int) -> None:
    if len(node.args) != nargs:
        fn_name = ast_utils.extract_function_name_or_none(node)
        raise exceptions.LatexifySyntaxError(
            f"Incorrect number of arguments for {fn_name}."
            f" expected: {nargs}, but got {len(node.args)}"
        )


_FUNCTION_EXPANDERS: dict[str, Callable[[FunctionExpander, ast.Call], ast.AST]] = {
    "atan2": _atan2_expander,
    "exp": _exp_expander,
    "exp2": _exp2_expander,
    "expm1": _expm1_expander,
    "hypot": _hypot_expander,
    "log1p": _log1p_expander,
    "pow": _pow_expander,
}



"""Frontend interfaces of latexify."""

from __future__ import annotations

from collections.abc import Callable
from typing import Any, overload

from latexify import ipython_wrappers


@overload
def algorithmic(
    fn: Callable[..., Any], **kwargs: Any
) -> ipython_wrappers.LatexifiedAlgorithm:
    ...


@overload
def algorithmic(
    **kwargs: Any,
) -> Callable[[Callable[..., Any]], ipython_wrappers.LatexifiedAlgorithm]:
    ...


def algorithmic(
    fn: Callable[..., Any] | None = None, **kwargs: Any
) -> (
    ipython_wrappers.LatexifiedAlgorithm
    | Callable[[Callable[..., Any]], ipython_wrappers.LatexifiedAlgorithm]
):
    """Attach LaTeX pretty-printing to the given function.

    This function works with or without specifying the target function as the
    positional argument. The following two syntaxes works similarly.
        - latexify.algorithmic(alg, **kwargs)
        - latexify.algorithmic(**kwargs)(alg)

    Args:
        fn: Callable to be wrapped.
        **kwargs: Arguments to control behavior. See also get_latex().

    Returns:
        - If `fn` is passed, returns the wrapped function.
        - Otherwise, returns the wrapper function with given settings.
    """
    if fn is not None:
        return ipython_wrappers.LatexifiedAlgorithm(fn, **kwargs)

    def wrapper(f):
        return ipython_wrappers.LatexifiedAlgorithm(f, **kwargs)

    return wrapper


@overload
def function(
    fn: Callable[..., Any], **kwargs: Any
) -> ipython_wrappers.LatexifiedFunction:
    ...


@overload
def function(
    **kwargs: Any,
) -> Callable[[Callable[..., Any]], ipython_wrappers.LatexifiedFunction]:
    ...


def function(
    fn: Callable[..., Any] | None = None, **kwargs: Any
) -> (
    ipython_wrappers.LatexifiedFunction
    | Callable[[Callable[..., Any]], ipython_wrappers.LatexifiedFunction]
):
    """Attach LaTeX pretty-printing to the given function.

    This function works with or without specifying the target function as the positional
    argument. The following two syntaxes works similarly.
        - latexify.function(fn, **kwargs)
        - latexify.function(**kwargs)(fn)

    Args:
        fn: Callable to be wrapped.
        **kwargs: Arguments to control behavior. See also get_latex().

    Returns:
        - If `fn` is passed, returns the wrapped function.
        - Otherwise, returns the wrapper function with given settings.
    """
    if fn is not None:
        return ipython_wrappers.LatexifiedFunction(fn, **kwargs)

    def wrapper(f):
        return ipython_wrappers.LatexifiedFunction(f, **kwargs)

    return wrapper


@overload
def expression(
    fn: Callable[..., Any], **kwargs: Any
) -> ipython_wrappers.LatexifiedFunction:
    ...


@overload
def expression(
    **kwargs: Any,
) -> Callable[[Callable[..., Any]], ipython_wrappers.LatexifiedFunction]:
    ...


def expression(
    fn: Callable[..., Any] | None = None, **kwargs: Any
) -> (
    ipython_wrappers.LatexifiedFunction
    | Callable[[Callable[..., Any]], ipython_wrappers.LatexifiedFunction]
):
    """Attach LaTeX pretty-printing to the given function.

    This function is a shortcut for `latexify.function` with the default parameter
    `use_signature=False`.
    """
    kwargs["use_signature"] = kwargs.get("use_signature", False)

    if fn is not None:
        return ipython_wrappers.LatexifiedFunction(fn, **kwargs)

    def wrapper(f):
        return ipython_wrappers.LatexifiedFunction(f, **kwargs)

    return wrapper



from typing import Any

from latexify import exceptions


def convert_constant(value: Any) -> str:
    """Helper to convert constant values to LaTeX.

    Args:
        value: A constant value.

    Returns:
        The LaTeX representation of `value`.
    """
    if value is None or isinstance(value, bool):
        return r"\mathrm{" + str(value) + "}"
    if isinstance(value, (int, float, complex)):
        # TODO(odashi): Support other symbols for the imaginary unit than j.
        return str(value)
    if isinstance(value, str):
        return r'\textrm{"' + value + '"}'
    if isinstance(value, bytes):
        return r"\textrm{" + str(value) + "}"
    if value is ...:
        return r"\cdots"
    raise exceptions.LatexifyNotSupportedError(
        f"Unrecognized constant: {type(value).__name__}"
    )



"""Wrapper objects for IPython to display output."""

from __future__ import annotations

import abc
from typing import Any, Callable, cast

from latexify import exceptions, generate_latex


class LatexifiedRepr(metaclass=abc.ABCMeta):
    """Object with LaTeX representation."""

    _fn: Callable[..., Any]

    def __init__(self, fn: Callable[..., Any], **kwargs) -> None:
        self._fn = fn

    @property
    def __doc__(self) -> str | None:
        return self._fn.__doc__

    @__doc__.setter
    def __doc__(self, val: str | None) -> None:
        self._fn.__doc__ = val

    @property
    def __name__(self) -> str:
        return self._fn.__name__

    @__name__.setter
    def __name__(self, val: str) -> None:
        self._fn.__name__ = val

    # After Python 3.7
    # @final
    def __call__(self, *args) -> Any:
        return self._fn(*args)

    @abc.abstractmethod
    def __str__(self) -> str:
        ...

    @abc.abstractmethod
    def _repr_html_(self) -> str | tuple[str, dict[str, Any]] | None:
        """IPython hook to display HTML visualization."""
        ...

    @abc.abstractmethod
    def _repr_latex_(self) -> str | tuple[str, dict[str, Any]] | None:
        """IPython hook to display LaTeX visualization."""
        ...


class LatexifiedAlgorithm(LatexifiedRepr):
    """Algorithm with latex representation."""

    _latex: str | None
    _error: str | None
    _ipython_latex: str | None
    _ipython_error: str | None

    def __init__(self, fn: Callable[..., Any], **kwargs) -> None:
        super().__init__(fn)

        try:
            self._latex = generate_latex.get_latex(
                fn, style=generate_latex.Style.ALGORITHMIC, **kwargs
            )
            self._error = None
        except exceptions.LatexifyError as e:
            self._latex = None
            self._error = f"{type(e).__name__}: {str(e)}"

        try:
            self._ipython_latex = generate_latex.get_latex(
                fn, style=generate_latex.Style.IPYTHON_ALGORITHMIC, **kwargs
            )
            self._ipython_error = None
        except exceptions.LatexifyError as e:
            self._ipython_latex = None
            self._ipython_error = f"{type(e).__name__}: {str(e)}"

    def __str__(self) -> str:
        return self._latex if self._latex is not None else cast(str, self._error)

    def _repr_html_(self) -> str | tuple[str, dict[str, Any]] | None:
        """IPython hook to display HTML visualization."""
        return (
            '<span style="color: red;">' + self._ipython_error + "</span>"
            if self._ipython_error is not None
            else None
        )

    def _repr_latex_(self) -> str | tuple[str, dict[str, Any]] | None:
        """IPython hook to display LaTeX visualization."""
        return (
            f"$ {self._ipython_latex} $"
            if self._ipython_latex is not None
            else self._ipython_error
        )


class LatexifiedFunction(LatexifiedRepr):
    """Function with latex representation."""

    _latex: str | None
    _error: str | None

    def __init__(self, fn: Callable[..., Any], **kwargs) -> None:
        super().__init__(fn, **kwargs)

        try:
            self._latex = self._latex = generate_latex.get_latex(
                fn, style=generate_latex.Style.FUNCTION, **kwargs
            )
            self._error = None
        except exceptions.LatexifyError as e:
            self._latex = None
            self._error = f"{type(e).__name__}: {str(e)}"

    def __str__(self) -> str:
        return self._latex if self._latex is not None else cast(str, self._error)

    def _repr_html_(self) -> str | tuple[str, dict[str, Any]] | None:
        """IPython hook to display HTML visualization."""
        return (
            '<span style="color: red;">' + self._error + "</span>"
            if self._error is not None
            else None
        )

    def _repr_latex_(self) -> str | tuple[str, dict[str, Any]] | None:
        """IPython hook to display LaTeX visualization."""
        return (
            rf"$$ \displaystyle {self._latex} $$"
            if self._latex is not None
            else self._error
        )



"""Package latexify.transformers."""

from latexify.transformers.assignment_reducer import AssignmentReducer
from latexify.transformers.aug_assign_replacer import AugAssignReplacer
from latexify.transformers.function_expander import FunctionExpander
from latexify.transformers.identifier_replacer import IdentifierReplacer
from latexify.transformers.prefix_trimmer import PrefixTrimmer

__all__ = [
    "AssignmentReducer",
    "AugAssignReplacer",
    "FunctionExpander",
    "IdentifierReplacer",
    "PrefixTrimmer",
]



"""NodeTransformer to trim unnecessary prefixes."""

from __future__ import annotations

import ast
import re

from latexify import ast_utils

_PREFIX_PATTERN = re.compile(r"^[A-Za-z_][A-Za-z0-9_]*(\.[A-Za-z_][A-Za-z0-9_]*)*$")


class PrefixTrimmer(ast.NodeTransformer):
    """NodeTransformer to trim unnecessary prefixes.

    This class investigates all Attribute subtrees, and replace them if the prefix of
    the attribute matches the given set of prefixes.
    Prefix is searched in the manner of leftmost longest matching.

    Example:
        def f(x):
            return math.sqrt(x)

        PrefixTrimmer({"math"}) will modify the AST of the function above to below:

        def f(x):
            return sqrt(x)
    """

    _prefixes: list[tuple[str, ...]]

    def __init__(self, prefixes: set[str]) -> None:
        """Initializer.

        Args:
            prefixes: Set of prefixes to be trimmed. Nested prefix is allowed too.
                Each value must follow one of the following formats:
                - A Python identifier, e.g., "math"
                - Python identifiers joined by periods, e.g., "numpy.random"
        """
        for p in prefixes:
            if not _PREFIX_PATTERN.match(p):
                raise ValueError(f"Invalid prefix: {p}")

        self._prefixes = [tuple(p.split(".")) for p in prefixes]

    def _get_prefix(self, node: ast.expr) -> tuple[str, ...] | None:
        """Helper to obtain nested prefix.

        Args:
            node: Node to investigate.

        Returns:
            The prefix tuple, or None if the node has unsupported syntax.
        """
        if isinstance(node, ast.Name):
            return (node.id,)

        if isinstance(node, ast.Attribute):
            parent = self._get_prefix(node.value)
            return parent + (node.attr,) if parent is not None else None

        return None

    def _make_attribute(self, prefix: tuple[str, ...], name: str) -> ast.expr:
        """Helper to generate a new Attribute or Name node.

        Args:
            prefix: List of prefixes.
            name: Attribute name.

        Returns:
            Name node if prefix == (), (possibly nested) Attribute node otherwise.
        """
        if not prefix:
            return ast_utils.make_name(name)

        parent = self._make_attribute(prefix[:-1], prefix[-1])
        return ast_utils.make_attribute(parent, name)

    def visit_Attribute(self, node: ast.Attribute) -> ast.expr:
        """Visit an Attribute node."""
        prefix = self._get_prefix(node.value)
        if prefix is None:
            return node

        # Performs leftmost longest match.
        # NOTE(odashi):
        # This implementation is very naive, but would work efficiently as long as the
        # number of patterns is small.
        matched_length = 0

        for p in self._prefixes:
            length = min(len(p), len(prefix))
            if prefix[:length] == p and length > matched_length:
                matched_length = length

        return self._make_attribute(prefix[matched_length:], node.attr)



"""Utility to convert identifiers."""

from __future__ import annotations

from latexify.codegen import expression_rules


class IdentifierConverter:
    r"""Converts Python identifiers to appropriate LaTeX expression.

    This converter applies following rules:
        - `foo` --> `\foo`, if `use_math_symbols == True` and the given identifier
          matches a supported math symbol name.
        - `x` --> `x`, if the given identifier is exactly 1 character (except `_`)
        - `foo_bar` --> `\mathrm{foo\_bar}`, otherwise.
    """

    _use_math_symbols: bool
    _use_mathrm: bool

    def __init__(self, *, use_math_symbols: bool, use_mathrm: bool = True) -> None:
        r"""Initializer.

        Args:
            use_math_symbols: Whether to convert identifiers with math symbol names to
                appropriate LaTeX command.
            use_mathrm: Whether to wrap the resulting expression by \mathrm, if
                applicable.
        """
        self._use_math_symbols = use_math_symbols
        self._use_mathrm = use_mathrm

    def convert(self, name: str) -> tuple[str, bool]:
        """Converts Python identifier to LaTeX expression.

        Args:
            name: Identifier name.

        Returns:
            Tuple of following values:
                - latex: Corresponding LaTeX expression.
                - is_single_character: Whether `latex` can be treated as a single
                    character or not.
        """
        if self._use_math_symbols and name in expression_rules.MATH_SYMBOLS:
            return "\\" + name, True

        if len(name) == 1 and name != "_":
            return name, True

        escaped = name.replace("_", r"\_")
        wrapped = rf"\mathrm{{{escaped}}}" if self._use_mathrm else escaped

        return wrapped, False



"""Exceptions used in Latexify."""


class LatexifyError(Exception):
    """Base class of all Latexify exceptions.

    Subclasses of this exception does not mean incorrect use of the library by the user
    at the interface level. These exceptions inform users that Latexify went into
    something wrong during processing the given functions.
    These exceptions are usually captured by the frontend functions (e.g., `with_latex`)
    to prevent destroying the entire program.
    Errors caused by wrong inputs should raise built-in exceptions.
    """

    ...


class LatexifyNotSupportedError(LatexifyError):
    """Some subtree in the AST is not supported by the current implementation.

    This error is raised when the library discovered incompatible syntaxes due to lack
    of the implementation. Possibly this error would be resolved in the future.
    """

    ...


class LatexifySyntaxError(LatexifyError):
    """Some subtree in the AST is not supported.

    This error is raised when the library discovered syntaxes that are not possible to
    be processed anymore. This error is essential, and wouldn't be resolved in the
    future.
    """

    ...



"""Codegen rules for single expressions."""

from __future__ import annotations

import ast
import dataclasses

# Precedences of operators for BoolOp, BinOp, UnaryOp, and Compare nodes.
# Note that this value affects only the appearance of surrounding parentheses for each
# expression, and does not affect the AST itself.
# See also:
# https://docs.python.org/3/reference/expressions.html#operator-precedence
_PRECEDENCES: dict[type[ast.AST], int] = {
    ast.Pow: 120,
    ast.UAdd: 110,
    ast.USub: 110,
    ast.Invert: 110,
    ast.Mult: 100,
    ast.MatMult: 100,
    ast.Div: 100,
    ast.FloorDiv: 100,
    ast.Mod: 100,
    ast.Add: 90,
    ast.Sub: 90,
    ast.LShift: 80,
    ast.RShift: 80,
    ast.BitAnd: 70,
    ast.BitXor: 60,
    ast.BitOr: 50,
    ast.In: 40,
    ast.NotIn: 40,
    ast.Is: 40,
    ast.IsNot: 40,
    ast.Lt: 40,
    ast.LtE: 40,
    ast.Gt: 40,
    ast.GtE: 40,
    ast.NotEq: 40,
    ast.Eq: 40,
    # NOTE(odashi):
    # We assume that the `not` operator has the same precedence with other unary
    # operators `+`, `-` and `~`, because the LaTeX counterpart $\lnot$ looks to have a
    # high precedence.
    # ast.Not: 30,
    ast.Not: 110,
    ast.And: 20,
    ast.Or: 10,
}

# NOTE(odashi):
# Function invocation is treated as a unary operator with a higher precedence.
# This ensures that the argument with a unary operator is wrapped:
#     exp(x) --> \exp x
#     exp(-x) --> \exp (-x)
#     -exp(x) --> - \exp x
_CALL_PRECEDENCE = _PRECEDENCES[ast.UAdd] + 1

_INF_PRECEDENCE = 1_000_000


def get_precedence(node: ast.AST) -> int:
    """Obtains the precedence of the subtree.

    Args:
        node: Subtree to investigate.

    Returns:
        If `node` is a subtree with some operator, returns the precedence of the
        operator. Otherwise, returns a number larger enough from other precedences.
    """
    if isinstance(node, ast.Call):
        return _CALL_PRECEDENCE

    if isinstance(node, (ast.BinOp, ast.UnaryOp, ast.BoolOp)):
        return _PRECEDENCES[type(node.op)]

    if isinstance(node, ast.Compare):
        # Compare operators have the same precedence. It is enough to check only the
        # first operator.
        return _PRECEDENCES[type(node.ops[0])]

    return _INF_PRECEDENCE


@dataclasses.dataclass(frozen=True)
class BinOperandRule:
    """Syntax rules for operands of BinOp."""

    # Whether to require wrapping operands by parentheses according to the precedence.
    wrap: bool = True

    # Whether to require wrapping operands by parentheses if the operand has the same
    # precedence with this operator.
    # This is used to control the behavior of non-associative operators.
    force: bool = False


@dataclasses.dataclass(frozen=True)
class BinOpRule:
    """Syntax rules for BinOp."""

    # Left/middle/right syntaxes to wrap operands.
    latex_left: str
    latex_middle: str
    latex_right: str

    # Operand rules.
    operand_left: BinOperandRule = dataclasses.field(default_factory=BinOperandRule)
    operand_right: BinOperandRule = dataclasses.field(default_factory=BinOperandRule)

    # Whether to assume the resulting syntax is wrapped by some bracket operators.
    # If True, the parent operator can avoid wrapping this operator by parentheses.
    is_wrapped: bool = False


BIN_OP_RULES: dict[type[ast.operator], BinOpRule] = {
    ast.Pow: BinOpRule(
        "",
        "^{",
        "}",
        operand_left=BinOperandRule(force=True),
        operand_right=BinOperandRule(wrap=False),
    ),
    ast.Mult: BinOpRule("", r" \cdot ", ""),
    ast.MatMult: BinOpRule("", r" \cdot ", ""),
    ast.Div: BinOpRule(
        r"\frac{",
        "}{",
        "}",
        operand_left=BinOperandRule(wrap=False),
        operand_right=BinOperandRule(wrap=False),
    ),
    ast.FloorDiv: BinOpRule(
        r"\left\lfloor\frac{",
        "}{",
        r"}\right\rfloor",
        operand_left=BinOperandRule(wrap=False),
        operand_right=BinOperandRule(wrap=False),
        is_wrapped=True,
    ),
    ast.Mod: BinOpRule(
        "", r" \mathbin{\%} ", "", operand_right=BinOperandRule(force=True)
    ),
    ast.Add: BinOpRule("", " + ", ""),
    ast.Sub: BinOpRule("", " - ", "", operand_right=BinOperandRule(force=True)),
    ast.LShift: BinOpRule("", r" \ll ", "", operand_right=BinOperandRule(force=True)),
    ast.RShift: BinOpRule("", r" \gg ", "", operand_right=BinOperandRule(force=True)),
    ast.BitAnd: BinOpRule("", r" \mathbin{\&} ", ""),
    ast.BitXor: BinOpRule("", r" \oplus ", ""),
    ast.BitOr: BinOpRule("", r" \mathbin{|} ", ""),
}

# Typeset for BinOp of sets.
SET_BIN_OP_RULES: dict[type[ast.operator], BinOpRule] = {
    **BIN_OP_RULES,
    ast.Sub: BinOpRule(
        "", r" \setminus ", "", operand_right=BinOperandRule(force=True)
    ),
    ast.BitAnd: BinOpRule("", r" \cap ", ""),
    ast.BitXor: BinOpRule("", r" \mathbin{\triangle} ", ""),
    ast.BitOr: BinOpRule("", r" \cup ", ""),
}

UNARY_OPS: dict[type[ast.unaryop], str] = {
    ast.Invert: r"\mathord{\sim} ",
    ast.UAdd: "+",  # Explicitly adds the $+$ operator.
    ast.USub: "-",
    ast.Not: r"\lnot ",
}

COMPARE_OPS: dict[type[ast.cmpop], str] = {
    ast.Eq: "=",
    ast.Gt: ">",
    ast.GtE: r"\ge",
    ast.In: r"\in",
    ast.Is: r"\equiv",
    ast.IsNot: r"\not\equiv",
    ast.Lt: "<",
    ast.LtE: r"\le",
    ast.NotEq: r"\ne",
    ast.NotIn: r"\notin",
}

# Typeset for Compare of sets.
SET_COMPARE_OPS: dict[type[ast.cmpop], str] = {
    **COMPARE_OPS,
    ast.Gt: r"\supset",
    ast.GtE: r"\supseteq",
    ast.Lt: r"\subset",
    ast.LtE: r"\subseteq",
}

BOOL_OPS: dict[type[ast.boolop], str] = {
    ast.And: r"\land",
    ast.Or: r"\lor",
}


@dataclasses.dataclass(frozen=True)
class FunctionRule:
    """Codegen rules for functions.

    Attributes:
        left: LaTeX expression concatenated to the left-hand side of the arguments.
        right: LaTeX expression concatenated to the right-hand side of the arguments.
        is_unary: Whether the function is treated as a unary operator or not.
        is_wrapped: Whether the resulting syntax is wrapped by brackets or not.
    """

    left: str
    right: str = ""
    is_unary: bool = False
    is_wrapped: bool = False


# name => left_syntax, right_syntax, is_wrapped
BUILTIN_FUNCS: dict[str, FunctionRule] = {
    "abs": FunctionRule(r"\mathopen{}\left|", r"\mathclose{}\right|", is_wrapped=True),
    "acos": FunctionRule(r"\arccos", is_unary=True),
    "acosh": FunctionRule(r"\mathrm{arcosh}", is_unary=True),
    "arccos": FunctionRule(r"\arccos", is_unary=True),
    "arccot": FunctionRule(r"\mathrm{arccot}", is_unary=True),
    "arccsc": FunctionRule(r"\mathrm{arccsc}", is_unary=True),
    "arcosh": FunctionRule(r"\mathrm{arcosh}", is_unary=True),
    "arcoth": FunctionRule(r"\mathrm{arcoth}", is_unary=True),
    "arcsec": FunctionRule(r"\mathrm{arcsec}", is_unary=True),
    "arcsch": FunctionRule(r"\mathrm{arcsch}", is_unary=True),
    "arcsin": FunctionRule(r"\arcsin", is_unary=True),
    "arctan": FunctionRule(r"\arctan", is_unary=True),
    "arsech": FunctionRule(r"\mathrm{arsech}", is_unary=True),
    "arsinh": FunctionRule(r"\mathrm{arsinh}", is_unary=True),
    "artanh": FunctionRule(r"\mathrm{artanh}", is_unary=True),
    "asin": FunctionRule(r"\arcsin", is_unary=True),
    "asinh": FunctionRule(r"\mathrm{arsinh}", is_unary=True),
    "atan": FunctionRule(r"\arctan", is_unary=True),
    "atanh": FunctionRule(r"\mathrm{artanh}", is_unary=True),
    "ceil": FunctionRule(
        r"\mathopen{}\left\lceil", r"\mathclose{}\right\rceil", is_wrapped=True
    ),
    "cos": FunctionRule(r"\cos", is_unary=True),
    "cosh": FunctionRule(r"\cosh", is_unary=True),
    "cot": FunctionRule(r"\cot", is_unary=True),
    "coth": FunctionRule(r"\coth", is_unary=True),
    "csc": FunctionRule(r"\csc", is_unary=True),
    "csch": FunctionRule(r"\mathrm{csch}", is_unary=True),
    "exp": FunctionRule(r"\exp", is_unary=True),
    "fabs": FunctionRule(r"\mathopen{}\left|", r"\mathclose{}\right|", is_wrapped=True),
    "factorial": FunctionRule("", "!", is_unary=True),
    "floor": FunctionRule(
        r"\mathopen{}\left\lfloor", r"\mathclose{}\right\rfloor", is_wrapped=True
    ),
    "fsum": FunctionRule(r"\sum", is_unary=True),
    "gamma": FunctionRule(r"\Gamma"),
    "log": FunctionRule(r"\log", is_unary=True),
    "log10": FunctionRule(r"\log_{10}", is_unary=True),
    "log2": FunctionRule(r"\log_2", is_unary=True),
    "prod": FunctionRule(r"\prod", is_unary=True),
    "sec": FunctionRule(r"\sec", is_unary=True),
    "sech": FunctionRule(r"\mathrm{sech}", is_unary=True),
    "sin": FunctionRule(r"\sin", is_unary=True),
    "sinh": FunctionRule(r"\sinh", is_unary=True),
    "sqrt": FunctionRule(r"\sqrt{", "}", is_wrapped=True),
    "sum": FunctionRule(r"\sum", is_unary=True),
    "tan": FunctionRule(r"\tan", is_unary=True),
    "tanh": FunctionRule(r"\tanh", is_unary=True),
}

MATH_SYMBOLS = {
    "aleph",
    "alpha",
    "beta",
    "beth",
    "chi",
    "daleth",
    "delta",
    "digamma",
    "epsilon",
    "eta",
    "gamma",
    "gimel",
    "hbar",
    "infty",
    "iota",
    "kappa",
    "lambda",
    "mu",
    "nabla",
    "nu",
    "omega",
    "phi",
    "pi",
    "psi",
    "rho",
    "sigma",
    "tau",
    "theta",
    "upsilon",
    "varepsilon",
    "varkappa",
    "varphi",
    "varpi",
    "varrho",
    "varsigma",
    "vartheta",
    "xi",
    "zeta",
    "Delta",
    "Gamma",
    "Lambda",
    "Omega",
    "Phi",
    "Pi",
    "Psi",
    "Sigma",
    "Theta",
    "Upsilon",
    "Xi",
}
